﻿<#
Download and install Android SDK
#>
Function Invoke-InteractiveProcess([String]$FilePath, [string[]]$ArgumentList)
{
    $startInfo = New-Object System.Diagnostics.ProcessStartInfo
    $startInfo.FileName = $FilePath
    $startInfo.Arguments = $ArgumentList
    $startInfo.UseShellExecute = $false
    $startInfo.RedirectStandardInput = $true
    $startInfo.RedirectStandardOutput = $true
    $startInfo.CreateNoWindow = $true
    $process = [System.Diagnostics.Process]::Start($startInfo)

    return $process
} 


# Android helper functions
Function Get-AndroidHomeFromRegistry
{

    # if ([Environment]::Is64BitOperatingSystem)
    # powershell v1 doesn't have is 64 bit flag.
    if ([environment]::GetEnvironmentVariable("ProgramFiles(x86)"))
    {
        $androidRegistryKey = "HKLM:\SOFTWARE\Wow6432Node\Android SDK Tools"
    }
    else
    {
        $androidRegistryKey = "HKLM:\SOFTWARE\Android SDK Tools"
    }

    if (Test-Path $androidRegistryKey)
    {
        $path = (Get-ItemProperty $androidRegistryKey Path).Path

        if (-not (Test-Path $path))
        {
            $path = $null
        }
    }

    return $path
}

Function Get-DependencyChecksLocation
{
    if ([environment]::GetEnvironmentVariable("ProgramFiles(x86)"))
    {
        $vsRegistryKey = "HKLM:\SOFTWARE\Wow6432Node\Microsoft\VisualStudio\14.0"
    }
    else
    {
        $vsRegistryKey = "HKLM:\SOFTWARE\Microsoft\VisualStudio\14.0"
    }

    if (Test-Path $vsRegistryKey)
    {
        $path = (Get-ItemProperty $vsRegistryKey ShellFolder).ShellFolder

        if (-not (Test-Path $path))
        {
            $path = $null
        }
    }

    return Join-Path $path "Apps\DependencyChecks.xml"
}

Function Format-PackageList($packages)
{
    "$([Environment]::Newline)  $($packages -join "$([Environment]::Newline)  ")$([Environment]::Newline)"
}

Function Get-AVDConfiguration([String]$FilePath)
{
    $result = @{}

    if (Test-Path $FilePath)
    {
        ForEach ($pair in (Get-Content $FilePath) -split "`n")
        {
            $pair = $pair -split "="

            if ($pair.Count -eq 2)
            {
                $result[$pair[0]] = $pair[1]
            }
        }
    }

    return $result
}

Function Set-AVDConfiguration([String]$FilePath, [Hashtable]$configHashtable)
{
    Remove-Item $FilePath

    ForEach ($key in $configHashtable.Keys | Sort-Object)
    {
        $keyValuePair = "$key=$($configHashtable[$key])"
        Add-Content $FilePath $keyValuePair
    }
}

Function Create-AVD([String]$Name, [String]$DeviceDefinition, [Hashtable]$ConfigurationOptions)
{
    if (-not $Name)
    {
        return
    }

    add-content $logFile "Creating AVD $Name"

    $processArgs = @("create avd",
                     "-n $Name",
                     "-t $currentApiLevel",
                     "-b default/armeabi-v7a")

    if ($DeviceDefinition)
    {
        $processArgs += "-d `"$DeviceDefinition`""
    }

    # Create the AVD
    $process = Invoke-InteractiveProcess `
                    -FilePath "$androidHome\tools\android.bat" `
                    -ArgumentList $processArgs

    while (-not $process.StandardOutput.EndOfStream)
    {
        # Answer 'no' to: Do you wish to create a custom hardware profile [no]
        $firstChar = $process.StandardOutput.Peek()
        if ($firstChar -eq 68)
        {
            $process.StandardInput.WriteLine("no")
        }

        $line = $process.StandardOutput.ReadLine()
        add-content $logFile "$line$([Environment]::NewLine)"
    }

    # needs to kill adb process here because if SecondaryInstaller is launched via ARP, adb
	# server will be launched and prevent ARP from installing other things.
        # need to have -erroraction silentlycontinue, otherwise we'll get an error when adb doesn't exist
    $processAdb = Get-Process adb -erroraction silentlycontinue 
	if ($processAdb)
	{
       $processAdb.Kill()
	}
    if (!$process.HasExited)
    {
	   $process.CloseMainWindow()
    }
    $process.Dispose()

    $avdConfigFile = "$env:HOMEPATH\.android\avd\$defaultAVDName.avd\config.ini"

    if (Test-Path $avdConfigFile)
    {
        $configHashtable = Get-AVDConfiguration $avdConfigFile

        foreach ($key in $ConfigurationOptions.Keys)
        {
            $configHashtable[$key] = $ConfigurationOptions[$key]
        }

        Set-AVDConfiguration $avdConfigFile $configHashtable

        add-content $logFile "Updating AVD $Name to the following hardware config:$([Environment]::NewLine)"

        Get-Content $avdConfigFile | add-content $logFile

        add-content $logFile "$([Environment]::NewLine)"
    }
}

Function Get-DeviceDefinitionNames()
{
    return (& $androidBatch list devices) `
        | Select-String `
            -AllMatches `
            -Pattern 'id: (?<id>\d+) or "(?<deviceName>[\s-_A-Za-z0-9\.]+)"' `
        | Select-Object -ExpandProperty Matches `
        | Foreach-Object {$_.Groups['deviceName'].Value}
}

Function Get-PreferredAVDDeviceDefinition()
{
    $nexusDevices = Get-DeviceDefinitionNames | Where-Object { $_.Contains("Nexus")}

    $deviceDefinition = ""
    $preferredDeviceDefinition = "Galaxy Nexus"

    if ($nexusDevices.Length -gt 0)
    {
        if ($nexusDevices -contains $preferredDeviceDefinition)
        {
            $deviceDefinition = $preferredDeviceDefinition
        }
        else
        {
            $deviceDefinition = $nexusDevices[0]
        }
    }

    return $deviceDefinition
}

$logFile = Join-Path $Env:temp AndroidSDK_SI.log
if ([IO.File]::Exists($logFile))
{
    Remove-Item $logFile
}

$androidHome = Get-AndroidHomeFromRegistry

# if androidHome doesn't exist then we don't have to select products.
if (!$androidHome)
{
   add-content $logFile "No Android SDK detected."
   Set-InstallResult -Succeeded $true
   exit
}

add-content $logFile "AndroidHome:"
add-content $logFile $androidHome

$androidBatch = Join-Path $androidHome "tools\android.bat"

$packageListingOutputRaw = & $androidBatch list sdk --all --extended
$anyAVDsFound = (& $androidBatch list avd) -match "Name: "

$packageRegex = 'id: \d+ or "(?<packageName>[-_A-Za-z0-9\.]+)"'
$availablePackages = [regex]::Matches($packageListingOutputRaw, $packageRegex) | ForEach { $_.Groups["packageName"].Value }

add-content $logFile "Android packages available:"
add-content $logFile "$(Format-PackageList $availablePackages)"

$dependencyChecksFilePath = Get-DependencyChecksLocation

Add-Content $logFile "using dependencychecks.xml: "
Add-Content $logFile $dependencyChecksFilePath

[xml]$xml = Get-Content $dependencyChecksFilePath

$document = $xml.DocumentElement

$androidDocObject = $document.ChildNodes | Where({$_.ProductName -eq "Android"})
	
$androidPackages = $androidDocObject.ChildNodes | Where({$_.ProductName -eq "AndroidPackage"})

if ($androidPackages)
{
    $desiredPackageNames = $androidPackages | Where({$_.GetAttribute("PackageName")})
    $desiredPackages = $desiredPackageNames | Select -ExpandProperty PackageName

    $currentApiLevel = ($desiredPackages -match "^android-")[0]
}

$packagesToInstall = $desiredPackages | Where { $availablePackages -contains $_ }
$packagesNotFound = $desiredPackages | Where { $availablePackages -notcontains $_ }

if ($packagesToInstall)
{
    add-content $logFile "Installing packages:"
    add-content $logFile "$(Format-PackageList $packagesToInstall)"

    $process = Invoke-InteractiveProcess `
                    -FilePath "$androidHome\tools\android.bat" `
                    -ArgumentList @("update sdk",
                                    "-u",
                                    "-a",
                                    "--filter $($packagesToInstall -join ",")")

    while (-not $process.StandardOutput.EndOfStream)
    {
        # Press 'y' when we see: Do you accept...
        $firstChar = $process.StandardOutput.Peek()
        if ($firstChar -eq 68)
        {
            $process.StandardInput.WriteLine("y")
        }

        $line = $process.StandardOutput.ReadLine()
        add-content $logFile "$line$([Environment]::NewLine)"

        if ($line -match 'Done. \d+ packages installed.')
        {
            break
        }
    }

    # needs to kill adb process here because if SecondaryInstaller is launched via ARP, adb
	# server will be launched and prevent ARP from installing other things.
        # need to have -erroraction silentlycontinue, otherwise we'll get an error when adb doesn't exist
    $processAdb = Get-Process adb -erroraction silentlycontinue 
    if ($processAdb)
    {
       $processAdb.Kill()
    }
    if (!$process.HasExited)
    {
	   $process.CloseMainWindow()
    }
    $process.Dispose()
}

# If there aren't any AVDs present on the system create one.
if (-not $anyAVDsFound)
{
    $deviceDefinition = Get-PreferredAVDDeviceDefinition

    if ($deviceDefinition -and $currentApiLevel)
    {
        $defaultAVDName = "AVD_$($deviceDefinition -replace ' ', '')_ToolsForApacheCordova"

        Create-AVD `
            -Name $defaultAVDName `
            -DeviceDefinition $deviceDefinition `
            -ConfigurationOptions @{"hw.keyboard"  = "yes"; 
                                    "hw.ramSize"  = "768"; 
                                    "vm.heapSize" = "64"}
    }
}

if ($packagesNotFound)
{
    add-content $logFile "WARNING: The following package(s) were not found:"
    add-content $logFile "$(Format-PackageList $packagesNotFound)"
}

Set-InstallResult -Succeeded $true

# SIG # Begin signature block
# MIIh6QYJKoZIhvcNAQcCoIIh2jCCIdYCAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCAuYhmrhyC4C9Pf
# IEYzsKoJmGb+VJ4eStrZgTEdgQfwNaCCC4MwggULMIID86ADAgECAhMzAAAAe6KB
# C4cRq+f8AAAAAAB7MA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTAwHhcNMTQxMDAxMTgwNjM4WhcNMTYwMTAxMTgwNjM4WjCBgzEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjENMAsGA1UECxMETU9Q
# UjEeMBwGA1UEAxMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMIIBIjANBgkqhkiG9w0B
# AQEFAAOCAQ8AMIIBCgKCAQEAvQcxxth2EB1eNeHpmIf/xHv5hBJHStPFJ2Zp22w2
# nNOhA3CXNR09AyJfXIEL50eYNUaSPv035IjaKWFTJMd9iXVt9WcwDtqZ40mvXhVE
# IyRFFgVmHcDac0t+/HxA4SPZKYlGHAvdNz8gd97MD7tIaeAXcx/qOxYqkhsujIJJ
# OKAqkfCh2w/3fWqRLgVrimZlJghs5DJOT5DL6br8nBYxqgjIsp4c8QsCsBZdlXxK
# m+7DAolGOmGJ8GwTtg1m3hcURkrTyXnyx34N2x22kH5gItwtcNicDAMnN6xc8ydI
# WviJIaR0Gj4Da8UiqTd9s2dNxuScWaOIuy/Ev/9DUf1zPQIDAQABo4IBejCCAXYw
# HwYDVR0lBBgwFgYIKwYBBQUHAwMGCisGAQQBgjc9BgEwHQYDVR0OBBYEFE8LE+6Q
# o6azvj+qlES88piRTzIdMFEGA1UdEQRKMEikRjBEMQ0wCwYDVQQLEwRNT1BSMTMw
# MQYDVQQFEyozODA3Nis2OGQyZjhiYi0wYTNiLTQwYjQtOWYyNC1lYjdlOTQxOWQx
# NjAwHwYDVR0jBBgwFoAU5vxfe7siAFjkck619CF0IzLm76wwVgYDVR0fBE8wTTBL
# oEmgR4ZFaHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVjdHMv
# TWljQ29kU2lnUENBXzIwMTAtMDctMDYuY3JsMFoGCCsGAQUFBwEBBE4wTDBKBggr
# BgEFBQcwAoY+aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWND
# b2RTaWdQQ0FfMjAxMC0wNy0wNi5jcnQwDAYDVR0TAQH/BAIwADANBgkqhkiG9w0B
# AQsFAAOCAQEAL1QbJw93iIl4XlutFtrDwEs8JT/kMaquI7SOYeIwgVH+SI71MCMZ
# TcveYIhzmIcEQz3gi+6gC28dVa6Y0GeuADG+vu3thnUKkko63nE4tGH3wDvhMRbx
# /L8wYXp13zaEV3R/HaYCiukoMfWMkN+qWdGepYUmp+PAXFTl3dVd/4oxMTdbLf4t
# lx1QqOyGSAyRrbJmPB57RWEJrGfhfYInTwr8Sy7UF/mVMxF7qCYBPJJD5pngvKsK
# TXNRW3YSyiaJ1TXHutME/dy4jipqbcZXUUA9qjRIICpPnoFxyk1VhBpKePeG96WJ
# MV0lBi+tzLVoWe5RSzxIKyI2qxzQho7SiTCCBnAwggRYoAMCAQICCmEMUkwAAAAA
# AAMwDQYJKoZIhvcNAQELBQAwgYgxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNo
# aW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29y
# cG9yYXRpb24xMjAwBgNVBAMTKU1pY3Jvc29mdCBSb290IENlcnRpZmljYXRlIEF1
# dGhvcml0eSAyMDEwMB4XDTEwMDcwNjIwNDAxN1oXDTI1MDcwNjIwNTAxN1owfjEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEoMCYGA1UEAxMfTWlj
# cm9zb2Z0IENvZGUgU2lnbmluZyBQQ0EgMjAxMDCCASIwDQYJKoZIhvcNAQEBBQAD
# ggEPADCCAQoCggEBAOkOZFB5Z7XE4/0JAEyelKz3VmjqRNjPxVhPqaV2fG1FutM5
# krSkHvn5ZYLkF9KP/UScCOhlk84sVYS/fQjjLiuoQSsYt6JLbklMaxUH3tHSwoke
# cZTNtX9LtK8I2MyI1msXlDqTziY/7Ob+NJhX1R1dSfayKi7VhbtZP/iQtCuDdMor
# sztG4/BGScEXZlTJHL0dxFViV3L4Z7klIDTeXaallV6rKIDN1bKe5QO1Y9OyFMjB
# yIomCll/B+z/Du2AEjVMEqa+Ulv1ptrgiwtId9aFR9UQucboqu6Lai0FXGDGtCpb
# nCMcX0XjGhQebzfLGTOAaolNo2pmY3iT1TDPlR8CAwEAAaOCAeMwggHfMBAGCSsG
# AQQBgjcVAQQDAgEAMB0GA1UdDgQWBBTm/F97uyIAWORyTrX0IXQjMubvrDAZBgkr
# BgEEAYI3FAIEDB4KAFMAdQBiAEMAQTALBgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUw
# AwEB/zAfBgNVHSMEGDAWgBTV9lbLj+iiXGJo0T2UkFvXzpoYxDBWBgNVHR8ETzBN
# MEugSaBHhkVodHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0
# cy9NaWNSb29DZXJBdXRfMjAxMC0wNi0yMy5jcmwwWgYIKwYBBQUHAQEETjBMMEoG
# CCsGAQUFBzAChj5odHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01p
# Y1Jvb0NlckF1dF8yMDEwLTA2LTIzLmNydDCBnQYDVR0gBIGVMIGSMIGPBgkrBgEE
# AYI3LgMwgYEwPQYIKwYBBQUHAgEWMWh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9Q
# S0kvZG9jcy9DUFMvZGVmYXVsdC5odG0wQAYIKwYBBQUHAgIwNB4yIB0ATABlAGcA
# YQBsAF8AUABvAGwAaQBjAHkAXwBTAHQAYQB0AGUAbQBlAG4AdAAuIB0wDQYJKoZI
# hvcNAQELBQADggIBABp071dPKXvEFoV4uFDTIvwJnayCl/g0/yosl5US5eS/z7+T
# yOM0qduBuNweAL7SNW+v5X95lXflAtTx69jNTh4bYaLCWiMa8IyoYlFFZwjjPzwe
# k/gwhRfIOUCm1w6zISnlpaFpjCKTzHSY56FHQ/JTrMAPMGl//tIlIG1vYdPfB9XZ
# cgAsaYZ2PVHbpjlIyTdhbQfdUxnLp9Zhwr/ig6sP4GubldZ9KFGwiUpRpJpsyLcf
# ShoOaanX3MF+0Ulwqratu3JHYxf6ptaipobsqBBEm2O2smmJBsdGhnoYP+jFHSHV
# e/kCIy3FQcu/HUzIFu+xnH/8IktJim4V46Z/dlvRU3mRhZ3V0ts9czXzPK5UslJH
# asCqE5XSjhHamWdeMoz7N4XR3HWFnIfGWleFwr/dDY+Mmy3rtO7PJ9O1Xmn6pBYE
# AackZ3PPTU+23gVWl3r36VJN9HcFT4XG2Avxju1CCdENduMjVngiJja+yrGMbqod
# 5IXaRzNij6TJkTNfcR5Ar5hlySLoQiElihwtYNk3iUGJKhYP12E8lGhgUu/WR5mg
# gEDuFYF3PpzgUxgaUB04lZseZjMTJzkXeIc2zk7DX7L1PUdTtuDl2wthPSrXkizO
# N1o+QEIxpB8QCMJWnL8kXVECnWp50hfT2sGUjgd7JXFEqwZq5tTG3yOalnXFMYIV
# vDCCFbgCAQEwgZUwfjELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24x
# EDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlv
# bjEoMCYGA1UEAxMfTWljcm9zb2Z0IENvZGUgU2lnbmluZyBQQ0EgMjAxMAITMwAA
# AHuigQuHEavn/AAAAAAAezANBglghkgBZQMEAgEFAKCBwjAZBgkqhkiG9w0BCQMx
# DAYKKwYBBAGCNwIBBDAcBgorBgEEAYI3AgELMQ4wDAYKKwYBBAGCNwIBFTAvBgkq
# hkiG9w0BCQQxIgQg76s295xxTRAOjG35nXX93TYWN045NvWAUh5e3LAbUMMwVgYK
# KwYBBAGCNwIBDDFIMEagLIAqAEEAbgBkAHIAbwBpAGQAUwBEAEsASQBuAHMAdABh
# AGwAbAAuAHAAcwAxoRaAFGh0dHA6Ly9taWNyb3NvZnQuY29tMA0GCSqGSIb3DQEB
# AQUABIIBAFtxekP/FcFiprzRLrfKSoYIrb4K/Ho9XFtlMQ5nP/9SLj4vyCXIK9Mu
# QJ8zB7LQ8dfgnpu+Yz4qO4Wz1RQpHwv1DchfabXvzyuCf7G+YVoefh1nxF6zH4JN
# sjOt8q3ScB/MEqp5Q8c0lM7950R9dmf1Sn+JRcakSakcfIppMZDAHkGyJXBfLYrk
# jLYFT0XrkB/mfVhxF41QAbJK+52bJ6Y+ufvlsGOw6dYLwotBtpbW+gaNhtSKb1AA
# 8XLRw4V7R20jwpa4V7pB0L3RrZKo/mtwr2HQV/Ir524/pimDsndZ3Nn0JO8u6oiS
# LaYmp74oiPEqaRe1zYZJrAUhtmAH99OhghMyMIITLgYKKwYBBAGCNwMDATGCEx4w
# ghMaBgkqhkiG9w0BBwKgghMLMIITBwIBAzEPMA0GCWCGSAFlAwQCAQUAMIIBNQYL
# KoZIhvcNAQkQAQSgggEkBIIBIDCCARwCAQEGCisGAQQBhFkKAwEwMTANBglghkgB
# ZQMEAgEFAAQgu5ftxupmLi9H71Q0RpWgGwzAAOVw4VkD0NaONtq28f4CBlRJG5/i
# LBgTMjAxNDExMTAxMDQxMzEuMzQ3WjAHAgEBgAIB9KCBsaSBrjCBqzELMAkGA1UE
# BhMCVVMxCzAJBgNVBAgTAldBMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
# aWNyb3NvZnQgQ29ycG9yYXRpb24xDTALBgNVBAsTBE1PUFIxJzAlBgNVBAsTHm5D
# aXBoZXIgRFNFIEVTTjpDMEY0LTMwODYtREVGODElMCMGA1UEAxMcTWljcm9zb2Z0
# IFRpbWUtU3RhbXAgU2VydmljZaCCDr0wggZxMIIEWaADAgECAgphCYEqAAAAAAAC
# MA0GCSqGSIb3DQEBCwUAMIGIMQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGlu
# Z3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBv
# cmF0aW9uMTIwMAYDVQQDEylNaWNyb3NvZnQgUm9vdCBDZXJ0aWZpY2F0ZSBBdXRo
# b3JpdHkgMjAxMDAeFw0xMDA3MDEyMTM2NTVaFw0yNTA3MDEyMTQ2NTVaMHwxCzAJ
# BgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25k
# MR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jv
# c29mdCBUaW1lLVN0YW1wIFBDQSAyMDEwMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8A
# MIIBCgKCAQEAqR0NvHcRijog7PwTl/X6f2mUa3RUENWlCgCChfvtfGhLLF/Fw+Vh
# wna3PmYrW/AVUycEMR9BGxqVHc4JE458YTBZsTBED/FgiIRUQwzXTbg4CLNC3ZOs
# 1nMwVyaCo0UN0Or1R4HNvyRgMlhgRvJYR4YyhB50YWeRX4FUsc+TTJLBxKZd0WET
# bijGGvmGgLvfYfxGwScdJGcSchohiq9LZIlQYrFd/XcfPfBXday9ikJNQFHRD5wG
# Pmd/9WbAA5ZEfu/QS/1u5ZrKsajyeioKMfDaTgaRtogINeh4HLDpmc085y9Euqf0
# 3GS9pAHBIAmTeM38vMDJRF1eFpwBBU8iTQIDAQABo4IB5jCCAeIwEAYJKwYBBAGC
# NxUBBAMCAQAwHQYDVR0OBBYEFNVjOlyKMZDzQ3t8RhvFM2hahW1VMBkGCSsGAQQB
# gjcUAgQMHgoAUwB1AGIAQwBBMAsGA1UdDwQEAwIBhjAPBgNVHRMBAf8EBTADAQH/
# MB8GA1UdIwQYMBaAFNX2VsuP6KJcYmjRPZSQW9fOmhjEMFYGA1UdHwRPME0wS6BJ
# oEeGRWh0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3RzL01p
# Y1Jvb0NlckF1dF8yMDEwLTA2LTIzLmNybDBaBggrBgEFBQcBAQROMEwwSgYIKwYB
# BQUHMAKGPmh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljUm9v
# Q2VyQXV0XzIwMTAtMDYtMjMuY3J0MIGgBgNVHSABAf8EgZUwgZIwgY8GCSsGAQQB
# gjcuAzCBgTA9BggrBgEFBQcCARYxaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL1BL
# SS9kb2NzL0NQUy9kZWZhdWx0Lmh0bTBABggrBgEFBQcCAjA0HjIgHQBMAGUAZwBh
# AGwAXwBQAG8AbABpAGMAeQBfAFMAdABhAHQAZQBtAGUAbgB0AC4gHTANBgkqhkiG
# 9w0BAQsFAAOCAgEAB+aIUQ3ixuCYP4FxAz2do6Ehb7Prpsz1Mb7PBeKp/vpXbRkw
# s8LFZslq3/Xn8Hi9x6ieJeP5vO1rVFcIK1GCRBL7uVOMzPRgEop2zEBAQZvcXBf/
# XPleFzWYJFZLdO9CEMivv3/Gf/I3fVo/HPKZeUqRUgCvOA8X9S95gWXZqbVr5MfO
# 9sp6AG9LMEQkIjzP7QOllo9ZKby2/QThcJ8ySif9Va8v/rbljjO7Yl+a21dA6fHO
# mWaQjP9qYn/dxUoLkSbiOewZSnFjnXshbcOco6I8+n99lmqQeKZt0uGc+R38ONiU
# 9MalCpaGpL2eGq4EQoO4tYCbIjggtSXlZOz39L9+Y1klD3ouOVd2onGqBooPiRa6
# YacRy5rYDkeagMXQzafQ732D8OE7cQnfXXSYIghh2rBQHm+98eEA3+cxB6STOvdl
# R3jo+KhIq/fecn5ha293qYHLpwmsObvsxsvYgrRyzR30uIUBHoD7G4kqVDmyW9rI
# DVWZeodzOwjmmC3qjeAzLhIp9cAvVCch98isTtoouLGp25ayp0Kiyc8ZQU3ghvkq
# mqMRZjDTu3QyS99je/WZii8bxyGvWbWu3EQ8l1Bx16HSxVXjad5XwdHeMMD9zOZN
# +w2/XU/pnR4ZOC+8z1gFLu8NoFA12u8JJxzVs341Hgi62jbb01+P3nSISRIwggTS
# MIIDuqADAgECAhMzAAAATDhdqKOTtc2aAAAAAABMMA0GCSqGSIb3DQEBCwUAMHwx
# CzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRt
# b25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1p
# Y3Jvc29mdCBUaW1lLVN0YW1wIFBDQSAyMDEwMB4XDTE0MDUyMzE3MjAwN1oXDTE1
# MDgyMzE3MjAwN1owgasxCzAJBgNVBAYTAlVTMQswCQYDVQQIEwJXQTEQMA4GA1UE
# BxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMQ0wCwYD
# VQQLEwRNT1BSMScwJQYDVQQLEx5uQ2lwaGVyIERTRSBFU046QzBGNC0zMDg2LURF
# RjgxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNlcnZpY2UwggEiMA0G
# CSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCaqcKeIbEYtsUydK8HgqJuV0lWoZ7y
# hVoaxreab65Yan0nLj/X+AjqX/cxGbYfUMGTDKA08QXLP2CzT6Ed9veqUsziq8x0
# VXYHCvlV6Pp2xSnZLUmZRnb6XECfB39DhGMSCK3eenrIQlPUV9Ok+R56s9toSnXd
# KH8DbcoyQbneCb/GG/sGc3QjtHSbPTxHMnm7LO5V4V5urVtypU2ttzOEMfizXYmG
# PoWAMNsSddIOm5Drz9A32adEyXfNDYsI1I6YgyHXw+lxCdSC7bbBQYJE8uuSR4/T
# 8yzKFOU/QOmvafpdBVaSRQi+nX7LYWGoFfie+85PhFNHwNzJ8dT91fwHAgMBAAGj
# ggEbMIIBFzAdBgNVHQ4EFgQUe5gigqCmKR4/mZCc71QCY357/8YwHwYDVR0jBBgw
# FoAU1WM6XIoxkPNDe3xGG8UzaFqFbVUwVgYDVR0fBE8wTTBLoEmgR4ZFaHR0cDov
# L2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVjdHMvTWljVGltU3RhUENB
# XzIwMTAtMDctMDEuY3JsMFoGCCsGAQUFBwEBBE4wTDBKBggrBgEFBQcwAoY+aHR0
# cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNUaW1TdGFQQ0FfMjAx
# MC0wNy0wMS5jcnQwDAYDVR0TAQH/BAIwADATBgNVHSUEDDAKBggrBgEFBQcDCDAN
# BgkqhkiG9w0BAQsFAAOCAQEAP8DCLjQIa+18e1IHZInYJghh2AvG80ikMji/mGQB
# XiF4IAQsvJu8QBMACtAawZQUbSXNqmuAM57XsPqIaCo19uYVo4AtuEAj+J8QdJbk
# 3LQgDgkjnnfGfYahWSM4fNxLik8Nnr9FgQoxDZJ2hqAHKHNe7cA2SC+r/vWTt5ZU
# sQfi0pxZ/7p2QGwjowTxBz3uSWMUe3I0xPyZEHLLOudMJcGFTow5kOKLeXvOkVob
# 90GrWm32Rso9et2/weeKleeZ4VPBOONVrVpuNwQBYnUgNoeRoFO5syReOtN4rMGn
# KEi3Y2gcxPZbWHDesjI2cF3rCYrNccJ80LZNRqnn4jZVkaGCA24wggJWAgEBMIHb
# oYGxpIGuMIGrMQswCQYDVQQGEwJVUzELMAkGA1UECBMCV0ExEDAOBgNVBAcTB1Jl
# ZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjENMAsGA1UECxME
# TU9QUjEnMCUGA1UECxMebkNpcGhlciBEU0UgRVNOOkMwRjQtMzA4Ni1ERUY4MSUw
# IwYDVQQDExxNaWNyb3NvZnQgVGltZS1TdGFtcCBTZXJ2aWNloiUKAQEwCQYFKw4D
# AhoFAAMVAKn+bYkfenG+uvmAAluZzy/zYFoeoIHCMIG/pIG8MIG5MQswCQYDVQQG
# EwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwG
# A1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMQ0wCwYDVQQLEwRNT1BSMScwJQYD
# VQQLEx5uQ2lwaGVyIE5UUyBFU046NTdGNi1DMUUwLTU1NEMxKzApBgNVBAMTIk1p
# Y3Jvc29mdCBUaW1lIFNvdXJjZSBNYXN0ZXIgQ2xvY2swDQYJKoZIhvcNAQEFBQAC
# BQDYCni6MCIYDzIwMTQxMTA5MjMzNTIyWhgPMjAxNDExMTAyMzM1MjJaMHQwOgYK
# KwYBBAGEWQoEATEsMCowCgIFANgKeLoCAQAwBwIBAAICBa0wBwIBAAICGRUwCgIF
# ANgLyjoCAQAwNgYKKwYBBAGEWQoEAjEoMCYwDAYKKwYBBAGEWQoDAaAKMAgCAQAC
# AxbjYKEKMAgCAQACAwehIDANBgkqhkiG9w0BAQUFAAOCAQEAfu5iE1Vr3ssZHx7t
# +n1DtQ5zyekCvPnyPe3n84J8aFxVQheTC7Alv7E/RVB2fG1p+5DXzP8tbJpfdUWM
# 2EywTv5oCTM7RBG6jAArGl/pYka6nj9bW7rTm5PSP+QM4k+xpUWFB1cCB37rZ0dN
# DLC3sQcUEFIv7LhozXMd8IV4cou9Ibl/hQ8AytqWwgOWq5p6FiRi4NULArYejsGe
# K5K60zgYiR3sE3Y+J2fNT2gfic1niI7yRGdBoR891T6l+Whrx7mHjfBjIpy48ldN
# kMYq2jpj6kzRDsmRXRTGmtQau0tdWLC8cl7JRtN8Mj1GTYz2z5OkO/T0hqArpklh
# 3ZJmETGCAvUwggLxAgEBMIGTMHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNo
# aW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29y
# cG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQSAyMDEw
# AhMzAAAATDhdqKOTtc2aAAAAAABMMA0GCWCGSAFlAwQCAQUAoIIBMjAaBgkqhkiG
# 9w0BCQMxDQYLKoZIhvcNAQkQAQQwLwYJKoZIhvcNAQkEMSIEIEPTwDQnyrm6R5Lb
# BEStd0dxpVAK0dWgnYCbIpjIoCuwMIHiBgsqhkiG9w0BCRACDDGB0jCBzzCBzDCB
# sQQUqf5tiR96cb66+YACW5nPL/NgWh4wgZgwgYCkfjB8MQswCQYDVQQGEwJVUzET
# MBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMV
# TWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQDEx1NaWNyb3NvZnQgVGltZS1T
# dGFtcCBQQ0EgMjAxMAITMwAAAEw4Xaijk7XNmgAAAAAATDAWBBRi5C3Ek32gxNhK
# 1i0QElTQ6LkmUDANBgkqhkiG9w0BAQsFAASCAQBZwT/hIwA4knIqehjkcY5op1yy
# YBBICFJuRjlyl+Kai79oduOf04mRBkiDrZW0I94+eP29bGcJrOpLXH0JBytFpiyV
# kRWa8Pa1fnJ2gAMcNllFGLzAeoZltsKHacy4dd5NTEFzdffx2x+tXxP+q7E6adVA
# aib6nSY4WGAVxEM6yW0SyowgWSx/f2FyL3/c3Ea+4KHRJt11p4vhEas2oinKQT73
# zR8zue7Dc2ITb7tEmYUaSxH7N/3ezezR7yo58AYyJYfCXX78eLgD35P2Z3xXKAot
# /5AEkRccxNEYTWL19scP+zitxiXfxDcvWxZKoiBs4ISiR8FiDMNkHnEHC+g1
# SIG # End signature block
